# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import math
import random
from collections import OrderedDict, namedtuple
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import librosa
import numpy as np
import torch

from nemo.collections.asr.parts.preprocessing.segment import AudioSegment, ChannelSelectorType
from nemo.collections.common.parts.preprocessing import collections
from nemo.collections.common.parts.utils import flatten
from nemo.core.classes import Dataset
from nemo.core.neural_types import AudioSignal, EncodedRepresentation, LengthsType, NeuralType
from nemo.utils import logging

__all__ = [
    'AudioToTargetDataset',
    'AudioToTargetWithReferenceDataset',
    'AudioToTargetWithEmbeddingDataset',
]


def _audio_collate_fn(batch: List[dict]) -> Tuple[torch.Tensor]:
    """Collate a batch of items returned by __getitem__.
    Examples for each signal are zero padded to the same length
    (batch_length), which is determined by the longest example.
    Lengths of the original signals are returned in the output.

    Args:
        batch: List of dictionaries. Each element of the list
            has the following format
            ```
            {
                'signal_0': 1D or 2D tensor,
                'signal_1': 1D or 2D tensor,
                ...
                'signal_N': 1D or 2D tensor,
            }
            ```
            1D tensors have shape (num_samples,) and 2D tensors
            have shape (num_channels, num_samples)

    Returns:
        A tuple containing signal tensor and signal length tensor (in samples)
        for each signal.
        The output has the following format:
        ```
        (signal_0, signal_0_length, signal_1, signal_1_length, ..., signal_N, signal_N_length)
        ```
        Note that the output format is obtained by interleaving signals and their length.
    """
    signals = batch[0].keys()

    batched = tuple()

    for signal in signals:
        signal_length = [b[signal].shape[-1] for b in batch]
        # Batch length is determined by the longest signal in the batch
        batch_length = max(signal_length)
        b_signal = []
        for s_len, b in zip(signal_length, batch):
            # check if padding is necessary
            if s_len < batch_length:
                if b[signal].ndim == 1:
                    # single-channel signal
                    pad = (0, batch_length - s_len)
                elif b[signal].ndim == 2:
                    # multi-channel signal
                    pad = (0, batch_length - s_len, 0, 0)
                else:
                    raise RuntimeError(
                        f'Signal {signal} has unsuported dimensions {signal.shape}. Currently, only 1D and 2D arrays are supported.'
                    )
                b[signal] = torch.nn.functional.pad(b[signal], pad)
            # append the current padded signal
            b_signal.append(b[signal])
        # (signal_batched, signal_length)
        batched += (torch.stack(b_signal), torch.tensor(signal_length, dtype=torch.int32))

    # Currently, outputs are expected to be in a tuple, where each element must correspond
    # to the output type in the OrderedDict returned by output_types.
    #
    # Therefore, we return batched signals by interleaving signals and their length:
    #   (signal_0, signal_0_length, signal_1, signal_1_length, ...)
    return batched


@dataclass
class SignalSetup:
    signals: List[str]  # signal names
    duration: Optional[Union[float, list]] = None  # duration for each signal
    channel_selectors: Optional[List[ChannelSelectorType]] = None  # channel selector for loading each signal


class ASRAudioProcessor:
    """Class that processes an example from Audio collection and returns
    a dictionary with prepared signals.

    For example, the output dictionary may be the following
    ```
    {
        'input_signal': input_signal_tensor,
        'target_signal': target_signal_tensor,
        'reference_signal': reference_signal_tensor,
        'embedding_vector': embedding_vector
    }
    ```
    Keys in the output dictionary are ordered with synchronous signals given first,
    followed by asynchronous signals and embedding.

    Args:
        sample_rate: sample rate used for all audio signals
        random_offset: If `True`, offset will be randomized when loading a subsegment
                       from a file.
        normalization_signal: Normalize all audio with a factor that ensures the signal
                    `example[normalization_signal]` in `process` is in range [-1, 1].
                    All other audio signals are scaled by the same factor. Default is
                    `None`, corresponding to no normalization.
    """

    def __init__(
        self,
        sample_rate: float,
        random_offset: bool,
        normalization_signal: Optional[str] = None,
        eps: float = 1e-8,
    ):
        self.sample_rate = sample_rate
        self.random_offset = random_offset
        self.normalization_signal = normalization_signal
        self.eps = eps

        self.sync_setup = None
        self.async_setup = None
        self.embedding_setup = None

    @property
    def sample_rate(self) -> float:
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value: float):
        if value <= 0:
            raise ValueError(f'Sample rate must be positive, received {value}')

        self._sample_rate = value

    @property
    def random_offset(self) -> bool:
        return self._random_offset

    @random_offset.setter
    def random_offset(self, value: bool):
        self._random_offset = value

    @property
    def sync_setup(self) -> SignalSetup:
        """Return the current setup for synchronous signals.

        Returns:
            A dataclass containing the list of signals, their
            duration and channel selectors.
        """
        return self._sync_setup

    @sync_setup.setter
    def sync_setup(self, value: Optional[SignalSetup]):
        """Setup signals to be loaded synchronously.

        Args:
            value: An instance of SignalSetup with the following fields
                - signals: list of signals (keys of example.audio_signals) which will be loaded
                           synchronously with the same start time and duration.
                - duration: Duration for each signal to be loaded.
                            If duration is set to None, the whole file will be loaded.
                - channel_selectors: A list of channel selector for each signal. If channel selector
                                     is None, all channels in the audio file will be loaded.
        """
        if value is None or isinstance(value, SignalSetup):
            self._sync_setup = value
        else:
            raise ValueError(f'Unexpected type {type(value)} for value {value}.')

    @property
    def async_setup(self) -> SignalSetup:
        """Return the current setup for asynchronous signals.

        Returns:
            A dataclass containing the list of signals, their
            duration and channel selectors.
        """
        return self._async_setup

    @async_setup.setter
    def async_setup(self, value: Optional[SignalSetup]):
        """Setup signals to be loaded asynchronously.

        Args:
        Args:
            value: An instance of SignalSetup with the following fields
                - signals: list of signals (keys of example.audio_signals) which will be loaded
                           asynchronously with signals possibly having different start and duration
                - duration: Duration for each signal to be loaded.
                            If duration is set to None, the whole file will be loaded.
                - channel_selectors: A list of channel selector for each signal. If channel selector
                                     is None, all channels in the audio file will be loaded.
        """
        if value is None or isinstance(value, SignalSetup):
            self._async_setup = value
        else:
            raise ValueError(f'Unexpected type {type(value)} for value {value}.')

    @property
    def embedding_setup(self) -> SignalSetup:
        """Setup signals corresponding to an embedding vector."""
        return self._embedding_setup

    @embedding_setup.setter
    def embedding_setup(self, value: SignalSetup):
        """Setup signals corresponding to an embedding vector.

        Args:
            value: An instance of SignalSetup with the following fields
                - signals: list of signals (keys of example.audio_signals) which will be loaded
                           as embedding vectors.
        """
        if value is None or isinstance(value, SignalSetup):
            self._embedding_setup = value
        else:
            raise ValueError(f'Unexpected type {type(value)} for value {value}.')

    def process(self, example: collections.Audio.OUTPUT_TYPE) -> Dict[str, torch.Tensor]:
        """Process an example from a collection of audio examples.

        Args:
            example: an example from Audio collection.

        Returns:
            An ordered dictionary of signals and their tensors.
            For example, the output dictionary may be the following
            ```
            {
                'input_signal': input_signal_tensor,
                'target_signal': target_signal_tensor,
                'reference_signal': reference_signal_tensor,
                'embedding_vector': embedding_vector
            }
            ```
            Keys in the output dictionary are ordered with synchronous signals given first,
            followed by asynchronous signals and embedding.
        """
        audio = self.load_audio(example=example)
        audio = self.process_audio(audio=audio)
        return audio

    def load_audio(self, example: collections.Audio.OUTPUT_TYPE) -> Dict[str, torch.Tensor]:
        """Given an example, load audio from `example.audio_files` and prepare
        the output dictionary.

        Args:
            example: An example from an audio collection

        Returns:
            An ordered dictionary of signals and their tensors.
            For example, the output dictionary may be the following
            ```
            {
                'input_signal': input_signal_tensor,
                'target_signal': target_signal_tensor,
                'reference_signal': reference_signal_tensor,
                'embedding_vector': embedding_vector
            }
            ```
            Keys in the output dictionary are ordered with synchronous signals given first,
            followed by asynchronous signals and embedding.
        """
        output = OrderedDict()

        if self.sync_setup is not None:
            # Load all signals with the same start and duration
            sync_signals = self.load_sync_signals(example)
            output.update(sync_signals)

        if self.async_setup is not None:
            # Load each signal independently
            async_signals = self.load_async_signals(example)
            output.update(async_signals)

        # Load embedding vector
        if self.embedding_setup is not None:
            embedding = self.load_embedding(example)
            output.update(embedding)

        if not output:
            raise RuntimeError('Output dictionary is empty. Please use `_setup` methods to setup signals to be loaded')

        return output

    def process_audio(self, audio: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Process audio signals available in the input dictionary.

        Args:
            audio: A dictionary containing loaded signals `signal: tensor`

        Returns:
            An ordered dictionary of signals and their tensors.
        """
        if self.normalization_signal:
            # Normalize all audio with a factor that ensures the normalization signal is in range [-1, 1].
            norm_scale = audio[self.normalization_signal].abs().max()

            # Do not normalize embeddings
            skip_signals = self.embedding_setup.signals if self.embedding_setup is not None else []

            # Normalize audio signals
            for signal in audio:
                if signal not in skip_signals:
                    # All audio signals are scaled by the same factor.
                    # This ensures that the relative level between signals is preserved.
                    audio[signal] = audio[signal] / (norm_scale + self.eps)

        return audio

    def load_sync_signals(self, example: collections.Audio.OUTPUT_TYPE) -> Dict[str, torch.Tensor]:
        """Load signals with the same start and duration.

        Args:
            example: an example from audio collection

        Returns:
            An ordered dictionary of signals and their tensors.
        """
        output = OrderedDict()
        sync_audio_files = [example.audio_files[s] for s in self.sync_setup.signals]

        sync_samples = self.get_samples_synchronized(
            audio_files=sync_audio_files,
            channel_selectors=self.sync_setup.channel_selectors,
            sample_rate=self.sample_rate,
            duration=self.sync_setup.duration,
            fixed_offset=example.offset,
            random_offset=self.random_offset,
        )

        for signal, samples in zip(self.sync_setup.signals, sync_samples):
            output[signal] = torch.tensor(samples)

        return output

    def load_async_signals(self, example: collections.Audio.OUTPUT_TYPE) -> Dict[str, torch.Tensor]:
        """Load each async signal independently, no constraints on starting
        from the same time.

        Args:
            example: an example from audio collection

        Returns:
            An ordered dictionary of signals and their tensors.
        """
        output = OrderedDict()
        for idx, signal in enumerate(self.async_setup.signals):
            samples = self.get_samples(
                audio_file=example.audio_files[signal],
                sample_rate=self.sample_rate,
                duration=self.async_setup.duration[idx],
                channel_selector=self.async_setup.channel_selectors[idx],
                fixed_offset=example.offset,
                random_offset=self.random_offset,
            )
            output[signal] = torch.tensor(samples)
        return output

    @classmethod
    def get_samples(
        cls,
        audio_file: str,
        sample_rate: int,
        duration: Optional[float] = None,
        channel_selector: ChannelSelectorType = None,
        fixed_offset: float = 0,
        random_offset: bool = False,
    ) -> np.ndarray:
        """Get samples from an audio file.
        For a single-channel signal, the output is shape (num_samples,).
        For a multi-channel signal, the output is shape (num_samples, num_channels).

        Args:
            audio_file: path to an audio file
            sample_rate: desired sample rate for output samples
            duration: Optional desired duration of output samples.
                    If `None`, the complete file will be loaded.
                    If set, a segment of `duration` seconds will be loaded.
            channel_selector: Optional channel selector, for selecting a subset of channels.
            fixed_offset: Optional fixed offset when loading samples.
            random_offset: If `True`, offset will be randomized when loading a short segment
                        from a file. The value is randomized between fixed_offset and
                        max_offset (set depending on the duration and fixed_offset).

        Returns:
            Numpy array with samples from audio file.
            The array has shape (num_samples,) for a single-channel signal
            or (num_channels, num_samples) for a multi-channel signal.
        """
        output = cls.get_samples_synchronized(
            audio_files=[audio_file],
            sample_rate=sample_rate,
            duration=duration,
            channel_selectors=[channel_selector],
            fixed_offset=fixed_offset,
            random_offset=random_offset,
        )

        return output[0]

    @classmethod
    def get_samples_synchronized(
        cls,
        audio_files: List[str],
        sample_rate: int,
        duration: Optional[float] = None,
        channel_selectors: Optional[List[ChannelSelectorType]] = None,
        fixed_offset: float = 0,
        random_offset: bool = False,
    ) -> List[np.ndarray]:
        """Get samples from multiple files with the same start and end point.

        Args:
            audio_files: list of paths to audio files
            sample_rate: desired sample rate for output samples
            duration: Optional desired duration of output samples.
                    If `None`, the complete files will be loaded.
                    If set, a segment of `duration` seconds will be loaded from
                    all files. Segment is synchronized across files, so that
                    start and end points are the same.
            channel_selectors: Optional channel selector for each signal, for selecting
                            a subset of channels.
            fixed_offset: Optional fixed offset when loading samples.
            random_offset: If `True`, offset will be randomized when loading a short segment
                        from a file. The value is randomized between fixed_offset and
                        max_offset (set depending on the duration and fixed_offset).

        Returns:
            List with the same size as `audio_files` but containing numpy arrays
            with samples from each audio file.
            Each array has shape (num_samples,) or (num_channels, num_samples), for single-
            or multi-channel signal, respectively.
            For example, if `audio_files = [path/to/file_1.wav, path/to/file_2.wav]`,
            the output will be a list `output = [samples_file_1, samples_file_2]`.
        """
        if channel_selectors is None:
            channel_selectors = [None] * len(audio_files)

        if duration is None:
            # Load complete files starting from a fixed offset
            offset = fixed_offset  # fixed offset
            num_samples = None  # no constrain on the number of samples

        else:
            # Fixed duration of the output
            audio_durations = cls.get_duration(audio_files)
            min_audio_duration = min(audio_durations)
            available_duration = min_audio_duration - fixed_offset

            if available_duration <= 0:
                raise ValueError(f'Fixed offset {fixed_offset}s is larger than shortest file {min_audio_duration}s.')

            if duration + fixed_offset > min_audio_duration:
                # The shortest file is shorter than the requested duration
                logging.debug(
                    f'Shortest file ({min_audio_duration}s) is less than the desired duration {duration}s + fixed offset {fixed_offset}s. Returned signals will be shortened to {available_duration} seconds.'
                )
                offset = fixed_offset
                duration = available_duration
            elif random_offset:
                # Randomize offset based on the shortest file
                max_offset = min_audio_duration - duration
                offset = random.uniform(fixed_offset, max_offset)
            else:
                # Fixed offset
                offset = fixed_offset

            # Fixed number of samples
            num_samples = math.floor(duration * sample_rate)

        output = []

        # Prepare segments
        for idx, audio_file in enumerate(audio_files):
            segment_samples = cls.get_samples_from_file(
                audio_file=audio_file,
                sample_rate=sample_rate,
                offset=offset,
                num_samples=num_samples,
                channel_selector=channel_selectors[idx],
            )
            output.append(segment_samples)

        return output

    @classmethod
    def get_samples_from_file(
        cls,
        audio_file: Union[str, List[str]],
        sample_rate: int,
        offset: float,
        num_samples: Optional[int] = None,
        channel_selector: Optional[ChannelSelectorType] = None,
    ) -> np.ndarray:
        """Get samples from a single or multiple files.
        If loading samples from multiple files, they will
        be concatenated along the channel dimension.

        Args:
            audio_file: path or a list of paths.
            sample_rate: sample rate of the loaded samples
            offset: fixed offset in seconds
            num_samples: Optional, number of samples to load.
                         If `None`, all available samples will be loaded.
            channel_selector: Select a subset of available channels.

        Returns:
            An array with shape (samples,) or (channels, samples)
        """
        if isinstance(audio_file, str):
            # Load samples from a single file
            segment_samples = cls.get_segment_from_file(
                audio_file=audio_file,
                sample_rate=sample_rate,
                offset=offset,
                num_samples=num_samples,
                channel_selector=channel_selector,
            )
        elif isinstance(audio_file, list):
            # Load samples from multiple files and form a multi-channel signal
            segment_samples = []
            for a_file in audio_file:
                a_file_samples = cls.get_segment_from_file(
                    audio_file=a_file,
                    sample_rate=sample_rate,
                    offset=offset,
                    num_samples=num_samples,
                    channel_selector=channel_selector,
                )
                segment_samples.append(a_file_samples)
            segment_samples = cls.list_to_multichannel(segment_samples)
        elif audio_file is None:
            # Support for inference, when the target signal is `None`
            segment_samples = []
        else:
            raise RuntimeError(f'Unexpected audio_file type {type(audio_file)}')
        return segment_samples

    @staticmethod
    def get_segment_from_file(
        audio_file: str,
        sample_rate: int,
        offset: float,
        num_samples: Optional[int] = None,
        channel_selector: Optional[ChannelSelectorType] = None,
    ) -> np.ndarray:
        """Get a segment of samples from a single audio file.

        Args:
            audio_file: path to an audio file
            sample_rate: sample rate of the loaded samples
            offset: fixed offset in seconds
            num_samples: Optional, number of samples to load.
                         If `None`, all available samples will be loaded.
            channel_selector: Select a subset of available channels.

        Returns:
           An array with shape (samples,) or (channels, samples)
        """
        if num_samples is None:
            segment = AudioSegment.from_file(
                audio_file=audio_file,
                target_sr=sample_rate,
                offset=offset,
                channel_selector=channel_selector,
            )

        else:
            segment = AudioSegment.segment_from_file(
                audio_file=audio_file,
                target_sr=sample_rate,
                n_segments=num_samples,
                offset=offset,
                channel_selector=channel_selector,
            )

        if segment.samples.ndim == 1:
            # Single-channel signal
            return segment.samples
        elif segment.samples.ndim == 2:
            # Use multi-channel format as (channels, samples)
            return segment.samples.T
        else:
            raise RuntimeError(f'Unexpected samples shape: {segment.samples.shape}')

    @staticmethod
    def list_to_multichannel(signal: Union[np.ndarray, List[np.ndarray]]) -> np.ndarray:
        """Convert a list of signals into a multi-channel signal by concatenating
        the elements of the list along the channel dimension.

        If input is not a list, it is returned unmodified.

        Args:
            signal: list of arrays

        Returns:
            Numpy array obtained by concatenating the elements of the list
            along the channel dimension (axis=0).
        """
        if not isinstance(signal, list):
            # Nothing to do there
            return signal
        elif len(signal) == 0:
            # Nothing to do, return as is
            return signal
        elif len(signal) == 1:
            # Nothing to concatenate, return the original format
            return signal[0]

        # If multiple signals are provided in a list, we concatenate them along the channel dimension
        if signal[0].ndim == 1:
            # Single-channel individual files
            mc_signal = np.stack(signal, axis=0)
        elif signal[0].ndim == 2:
            # Multi-channel individual files
            mc_signal = np.concatenate(signal, axis=0)
        else:
            raise RuntimeError(f'Unexpected target with {signal[0].ndim} dimensions.')

        return mc_signal

    @staticmethod
    def get_duration(audio_files: List[str]) -> List[float]:
        """Get duration for each audio file in `audio_files`.

        Args:
            audio_files: list of paths to audio files

        Returns:
            List of durations in seconds.
        """
        duration = [librosa.get_duration(path=f) for f in flatten(audio_files)]
        return duration

    def load_embedding(self, example: collections.Audio.OUTPUT_TYPE) -> Dict[str, torch.Tensor]:
        """Given an example, load embedding from `example.audio_files[embedding]`
        and return it in a dictionary.

        Args:
            example: An example from audio collection

        Returns:
            An dictionary of embedding keys and their tensors.
        """
        output = OrderedDict()
        for idx, signal in enumerate(self.embedding_setup.signals):
            embedding_file = example.audio_files[signal]
            embedding = self.load_embedding_vector(embedding_file)
            output[signal] = torch.tensor(embedding)
        return output

    @staticmethod
    def load_embedding_vector(filepath: str) -> np.ndarray:
        """Load an embedding vector from a file.

        Args:
            filepath: path to a file storing a vector.
                    Currently, it is assumed the file is a npy file.

        Returns:
            Array loaded from filepath.
        """
        if filepath.endswith('.npy'):
            with open(filepath, 'rb') as f:
                embedding = np.load(f)
        else:
            raise RuntimeError(f'Unknown embedding file format in file: {filepath}')

        return embedding


class BaseAudioDataset(Dataset):
    """Base class of audio datasets, providing common functionality
    for other audio datasets.

    Args:
        collection: Collection of audio examples prepared from manifest files.
        audio_processor: Used to process every example from the collection.
                         A callable with `process` method. For reference,
                         please check ASRAudioProcessor.
    """

    @property
    @abc.abstractmethod
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Returns definitions of module output ports."""

    def __init__(self, collection: collections.Audio, audio_processor: Callable, output_type: Type[namedtuple]):
        """Instantiates an audio dataset."""
        super().__init__()

        self.collection = collection
        self.audio_processor = audio_processor
        self.output_type = output_type

    def num_channels(self, signal_key) -> int:
        """Returns the number of channels for a particular signal in
        items prepared by this dictionary.

        More specifically, this will get the tensor from the first
        item in the dataset, check if it's a one- or two-dimensional
        tensor, and return the number of channels based on the size
        of the first axis (shape[0]).

        NOTE:
        This assumes that all examples have the same number of channels.

        Args:
            signal_key: string, used to select a signal from the dictionary
                        output by __getitem__

        Returns:
            Number of channels for the selected signal.
        """
        # Assumption: whole dataset has the same number of channels
        item = self.__getitem__(0)

        if item[signal_key].ndim == 1:
            return 1
        elif item[signal_key].ndim == 2:
            return item[signal_key].shape[0]
        else:
            raise RuntimeError(
                f'Unexpected number of dimension for signal {signal_key} with shape {item[signal_key].shape}'
            )

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        """Return a single example from the dataset.

        Args:
            index: integer index of an example in the collection

        Returns:
            Dictionary providing mapping from signal to its tensor.
            For example:
            ```
            {
                'input_signal': input_signal_tensor,
                'target_signal': target_signal_tensor,
            }
            ```
        """
        example = self.collection[index]
        output = self.audio_processor.process(example=example)

        return output

    def __len__(self) -> int:
        """Return the number of examples in the dataset."""
        return len(self.collection)

    def _collate_fn(self, batch) -> Tuple[torch.Tensor]:
        """Collate items in a batch."""
        return self.output_type(*_audio_collate_fn(batch))


AudioToTargetExample = namedtuple(
    typename='AudioToTargetExample', field_names='input_signal input_length target_signal target_length'
)


class AudioToTargetDataset(BaseAudioDataset):
    """A dataset for audio-to-audio tasks where the goal is to use
    an input signal to recover the corresponding target signal.

    Each line of the manifest file is expected to have the following format:

    .. code-block:: json

        {"input_key": "path/to/input.wav", "target_key": "path/to/target.wav", "duration": "duration_in_seconds"}

    Additionally, multiple audio files may be provided for each key in the manifest, for example,

    .. code-block:: json

        {"input_key": "path/to/input.wav", "target_key": ["path/to/path_to_target_ch0.wav", "path/to/path_to_target_ch1.wav"], "duration": "duration_in_seconds"}

    Keys for input and target signals can be configured in the constructor (`input_key` and `target_key`).

    Args:
        manifest_filepath: Path to manifest file in a format described above.
        sample_rate: Sample rate for loaded audio signals.
        input_key: Key pointing to input audio files in the manifest
        target_key: Key pointing to target audio files in manifest
        audio_duration: Optional duration of each item returned by __getitem__.
                        If `None`, complete audio will be loaded.
                        If set, a random subsegment will be loaded synchronously from
                        target and audio, i.e., with the same start and end point.
        random_offset: If `True`, offset will be randomized when loading a subsegment
                       from a file.
        max_duration: If audio exceeds this length, do not include in dataset.
        min_duration: If audio is less than this length, do not include in dataset.
        max_utts: Limit number of utterances.
        input_channel_selector: Optional, select subset of channels from each input audio file.
                                If `None`, all channels will be loaded.
        target_channel_selector: Optional, select subset of channels from each input audio file.
                                 If `None`, all channels will be loaded.
        normalization_signal: Normalize audio signals with a scale that ensures the normalization signal is in range [-1, 1].
                              All audio signals are scaled by the same factor. Supported values are `None` (no normalization),
                              'input_signal', 'target_signal'.
    """

    def __init__(
        self,
        manifest_filepath: str,
        sample_rate: int,
        input_key: str,
        target_key: str,
        audio_duration: Optional[float] = None,
        random_offset: bool = False,
        max_duration: Optional[float] = None,
        min_duration: Optional[float] = None,
        max_utts: Optional[int] = None,
        input_channel_selector: Optional[int] = None,
        target_channel_selector: Optional[int] = None,
        normalization_signal: Optional[str] = None,
    ):
        audio_to_manifest_key = {
            'input_signal': input_key,
            'target_signal': target_key,
        }

        collection = collections.AudioCollection(
            manifest_files=manifest_filepath,
            audio_to_manifest_key=audio_to_manifest_key,
            min_duration=min_duration,
            max_duration=max_duration,
            max_number=max_utts,
        )

        audio_processor = ASRAudioProcessor(
            sample_rate=sample_rate,
            random_offset=random_offset,
            normalization_signal=normalization_signal,
        )
        audio_processor.sync_setup = SignalSetup(
            signals=['input_signal', 'target_signal'],
            duration=audio_duration,
            channel_selectors=[input_channel_selector, target_channel_selector],
        )

        super().__init__(collection=collection, audio_processor=audio_processor, output_type=AudioToTargetExample)

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Returns definitions of module output ports.

        Returns:
            OrderedDict: Dictionary containing the following items:
                input_signal:
                    Batched single- or multi-channel input audio signal
                input_length:
                    Batched original length of each input signal
                target_signal:
                    Batched single- or multi-channel target audio signal
                target_length:
                    Batched original length of each target signal
        """
        sc_audio_type = NeuralType(('B', 'T'), AudioSignal())
        mc_audio_type = NeuralType(('B', 'C', 'T'), AudioSignal())

        return OrderedDict(
            input_signal=sc_audio_type if self.num_channels('input_signal') == 1 else mc_audio_type,
            input_length=NeuralType(('B',), LengthsType()),
            target_signal=sc_audio_type if self.num_channels('target_signal') == 1 else mc_audio_type,
            target_length=NeuralType(('B',), LengthsType()),
        )


AudioToTargetWithReferenceExample = namedtuple(
    typename='AudioToTargetWithReferenceExample',
    field_names='input_signal input_length target_signal target_length reference_signal reference_length',
)


class AudioToTargetWithReferenceDataset(BaseAudioDataset):
    """A dataset for audio-to-audio tasks where the goal is to use
    an input signal to recover the corresponding target signal and an
    additional reference signal is available.

    This can be used, for example, when a reference signal is
    available from
    - enrollment utterance for the target signal
    - echo reference from playback
    - reference from another sensor that correlates with the target signal

    Each line of the manifest file is expected to have the following format

    .. code-block:: json

        {"input_key": "path/to/input.wav", "target_key": "path/to/path_to_target.wav", "reference_key": "path/to/path_to_reference.wav", "duration": "duration_in_seconds"}

    Keys for input, target and reference signals can be configured in the constructor.

    Args:
        manifest_filepath: Path to manifest file in a format described above.
        sample_rate: Sample rate for loaded audio signals.
        input_key: Key pointing to input audio files in the manifest
        target_key: Key pointing to target audio files in manifest
        reference_key: Key pointing to reference audio files in manifest
        audio_duration: Optional duration of each item returned by __getitem__.
                        If `None`, complete audio will be loaded.
                        If set, a random subsegment will be loaded synchronously from
                        target and audio, i.e., with the same start and end point.
        random_offset: If `True`, offset will be randomized when loading a subsegment
                       from a file.
        max_duration: If audio exceeds this length, do not include in dataset.
        min_duration: If audio is less than this length, do not include in dataset.
        max_utts: Limit number of utterances.
        input_channel_selector: Optional, select subset of channels from each input audio file.
                                If `None`, all channels will be loaded.
        target_channel_selector: Optional, select subset of channels from each input audio file.
                                 If `None`, all channels will be loaded.
        reference_channel_selector: Optional, select subset of channels from each input audio file.
                                    If `None`, all channels will be loaded.
        reference_is_synchronized: If True, it is assumed that the reference signal is synchronized
                                   with the input signal, so the same subsegment will be loaded as for
                                   input and target. If False, reference signal will be loaded independently
                                   from input and target.
        reference_duration: Optional, can be used to set a fixed duration of the reference utterance. If `None`,
                            complete audio file will be loaded.
        normalization_signal: Normalize audio signals with a scale that ensures the normalization signal is in range [-1, 1].
                              All audio signals are scaled by the same factor. Supported values are `None` (no normalization),
                              'input_signal', 'target_signal', 'reference_signal'.
    """

    def __init__(
        self,
        manifest_filepath: str,
        sample_rate: int,
        input_key: str,
        target_key: str,
        reference_key: str,
        audio_duration: Optional[float] = None,
        random_offset: bool = False,
        max_duration: Optional[float] = None,
        min_duration: Optional[float] = None,
        max_utts: Optional[int] = None,
        input_channel_selector: Optional[int] = None,
        target_channel_selector: Optional[int] = None,
        reference_channel_selector: Optional[int] = None,
        reference_is_synchronized: bool = True,
        reference_duration: Optional[float] = None,
        normalization_signal: Optional[str] = None,
    ):
        audio_to_manifest_key = {
            'input_signal': input_key,
            'target_signal': target_key,
            'reference_signal': reference_key,
        }

        collection = collections.AudioCollection(
            manifest_files=manifest_filepath,
            audio_to_manifest_key=audio_to_manifest_key,
            min_duration=min_duration,
            max_duration=max_duration,
            max_number=max_utts,
        )

        audio_processor = ASRAudioProcessor(
            sample_rate=sample_rate,
            random_offset=random_offset,
            normalization_signal=normalization_signal,
        )

        if reference_is_synchronized:
            audio_processor.sync_setup = SignalSetup(
                signals=['input_signal', 'target_signal', 'reference_signal'],
                duration=audio_duration,
                channel_selectors=[input_channel_selector, target_channel_selector, reference_channel_selector],
            )
        else:
            audio_processor.sync_setup = SignalSetup(
                signals=['input_signal', 'target_signal'],
                duration=audio_duration,
                channel_selectors=[input_channel_selector, target_channel_selector],
            )
            audio_processor.async_setup = SignalSetup(
                signals=['reference_signal'],
                duration=[reference_duration],
                channel_selectors=[reference_channel_selector],
            )

        super().__init__(
            collection=collection, audio_processor=audio_processor, output_type=AudioToTargetWithReferenceExample
        )

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Returns definitions of module output ports.

        Returns:
            OrderedDict: Dictionary containing the following items:
                input_signal:
                    Batched single- or multi-channel input audio signal
                input_length:
                    Batched original length of each input signal
                target_signal:
                    Batched single- or multi-channel target audio signal
                target_length:
                    Batched original length of each target signal
                reference_signal:
                    Batched single- or multi-channel reference audio signal
                reference_length:
                    Batched original length of each reference signal
        """
        sc_audio_type = NeuralType(('B', 'T'), AudioSignal())
        mc_audio_type = NeuralType(('B', 'C', 'T'), AudioSignal())

        return OrderedDict(
            input_signal=sc_audio_type if self.num_channels('input_signal') == 1 else mc_audio_type,
            input_length=NeuralType(('B',), LengthsType()),
            target_signal=sc_audio_type if self.num_channels('target_signal') == 1 else mc_audio_type,
            target_length=NeuralType(('B',), LengthsType()),
            reference_signal=sc_audio_type if self.num_channels('reference_signal') == 1 else mc_audio_type,
            reference_length=NeuralType(('B',), LengthsType()),
        )


AudioToTargetWithEmbeddingExample = namedtuple(
    typename='AudioToTargetWithEmbeddingExample',
    field_names='input_signal input_length target_signal target_length embedding_vector embedding_length',
)


class AudioToTargetWithEmbeddingDataset(BaseAudioDataset):
    """A dataset for audio-to-audio tasks where the goal is to use
    an input signal to recover the corresponding target signal and an
    additional embedding signal. It is assumed that the embedding
    is in a form of a vector.

    Each line of the manifest file is expected to have the following format

    .. code-block:: json

        {"input_key": "path/to/input.wav", "target_key": "path/to/path_to_target.wav", "embedding_key": "path/to/path_to_reference.npy", "duration": "duration_in_seconds"}

    Keys for input, target and embedding signals can be configured in the constructor.

    Args:
        manifest_filepath: Path to manifest file in a format described above.
        sample_rate: Sample rate for loaded audio signals.
        input_key: Key pointing to input audio files in the manifest
        target_key: Key pointing to target audio files in manifest
        embedding_key: Key pointing to embedding files in manifest
        audio_duration: Optional duration of each item returned by __getitem__.
                        If `None`, complete audio will be loaded.
                        If set, a random subsegment will be loaded synchronously from
                        target and audio, i.e., with the same start and end point.
        random_offset: If `True`, offset will be randomized when loading a subsegment
                       from a file.
        max_duration: If audio exceeds this length, do not include in dataset.
        min_duration: If audio is less than this length, do not include in dataset.
        max_utts: Limit number of utterances.
        input_channel_selector: Optional, select subset of channels from each input audio file.
                                If `None`, all channels will be loaded.
        target_channel_selector: Optional, select subset of channels from each input audio file.
                                 If `None`, all channels will be loaded.
        normalization_signal: Normalize audio signals with a scale that ensures the normalization signal is in range [-1, 1].
                              All audio signals are scaled by the same factor. Supported values are `None` (no normalization),
                              'input_signal', 'target_signal'.
    """

    def __init__(
        self,
        manifest_filepath: str,
        sample_rate: int,
        input_key: str,
        target_key: str,
        embedding_key: str,
        audio_duration: Optional[float] = None,
        random_offset: bool = False,
        max_duration: Optional[float] = None,
        min_duration: Optional[float] = None,
        max_utts: Optional[int] = None,
        input_channel_selector: Optional[int] = None,
        target_channel_selector: Optional[int] = None,
        normalization_signal: Optional[str] = None,
    ):
        audio_to_manifest_key = {
            'input_signal': input_key,
            'target_signal': target_key,
            'embedding_vector': embedding_key,
        }

        collection = collections.AudioCollection(
            manifest_files=manifest_filepath,
            audio_to_manifest_key=audio_to_manifest_key,
            min_duration=min_duration,
            max_duration=max_duration,
            max_number=max_utts,
        )

        audio_processor = ASRAudioProcessor(
            sample_rate=sample_rate,
            random_offset=random_offset,
            normalization_signal=normalization_signal,
        )
        audio_processor.sync_setup = SignalSetup(
            signals=['input_signal', 'target_signal'],
            duration=audio_duration,
            channel_selectors=[input_channel_selector, target_channel_selector],
        )
        audio_processor.embedding_setup = SignalSetup(signals=['embedding_vector'])

        super().__init__(
            collection=collection, audio_processor=audio_processor, output_type=AudioToTargetWithEmbeddingExample
        )

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Returns definitions of module output ports.

        Returns:
            OrderedDict: Dictionary containing the following items:
                input_signal:
                    Batched single- or multi-channel input audio signal
                input_length:
                    Batched original length of each input signal
                target_signal:
                    Batched single- or multi-channel target audio signal
                target_length:
                    Batched original length of each target signal
                embedding_vector:
                    Batched embedded vector format
                embedding_length:
                    Batched original length of each embedding vector
        """
        sc_audio_type = NeuralType(('B', 'T'), AudioSignal())
        mc_audio_type = NeuralType(('B', 'C', 'T'), AudioSignal())

        return OrderedDict(
            input_signal=sc_audio_type if self.num_channels('input_signal') == 1 else mc_audio_type,
            input_length=NeuralType(('B',), LengthsType()),
            target_signal=sc_audio_type if self.num_channels('target_signal') == 1 else mc_audio_type,
            target_length=NeuralType(('B',), LengthsType()),
            embedding_vector=NeuralType(('B', 'D'), EncodedRepresentation()),
            embedding_length=NeuralType(('B',), LengthsType()),
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Union

import numpy as np
from kaldialign import align

from nemo.collections.asr.parts.context_biasing.ctc_based_word_spotter import WSHyp
from nemo.collections.asr.parts.utils import rnnt_utils
from nemo.collections.asr.parts.utils.manifest_utils import read_manifest
from nemo.utils import logging


def merge_alignment_with_ws_hyps(
    candidate: Union[np.ndarray, rnnt_utils.Hypothesis],
    asr_model,
    cb_results: List[WSHyp],
    decoder_type: str = "ctc",
    intersection_threshold: float = 30.0,
    blank_idx: int = 0,
    print_stats: bool = False,
    bow: str = "▁",
) -> tuple[str, str]:
    """
    Merge context biasing predictions with ctc/rnnt word-level alignment.
    Words from alignment will be replaced by spotted words if intersection between them is greater than threshold.

    Args:
        candidate: argmax predictions per frame (for ctc) or rnnt hypothesis (for rnnt)
        asr_model: ctc or hybrid transducer-ctc model
        cb_results: list of context biasing predictions (spotted words)
        decoder_type: ctc or rnnt
        intersection_threshold: threshold for intersection between spotted word and word from alignment (in percentage)
        blank_idx: blank index for ctc/rnnt decoding
        print_stats: if True, print word alignment and spotted words
        bow: symbol for begin of word (bow) in BPE tokenizer
    Returns:
        boosted_text: final text with context biasing predictions
    """

    # step 1: get token-level alignment [frame, token]
    if decoder_type == "ctc":
        alignment_tokens = []
        prev_token = None
        for idx, token in enumerate(candidate):
            if token != blank_idx:
                if token == prev_token:
                    alignment_tokens[-1] = [idx, asr_model.tokenizer.ids_to_tokens([int(token)])[0]]
                else:
                    alignment_tokens.append([idx, asr_model.tokenizer.ids_to_tokens([int(token)])[0]])
            prev_token = token

    elif decoder_type == "rnnt":
        alignment_tokens = []
        if not isinstance(candidate.y_sequence, list):
            candidate.y_sequence = candidate.y_sequence.tolist()
        tokens = asr_model.tokenizer.ids_to_tokens(candidate.y_sequence)
        for idx, token in enumerate(tokens):
            # bow symbol may be predicted separately from token
            if token == bow:
                if idx + 1 < len(tokens) and not tokens[idx + 1].startswith(bow):
                    tokens[idx + 1] = bow + tokens[idx + 1]
                    continue
            alignment_tokens.append([candidate.timestamp[idx].item(), token])
    else:
        raise ValueError(f"decoder_type {decoder_type} is not supported")

    if not alignment_tokens:
        # ctc/rnnt decoding results are empty, return context biasing results only
        return " ".join([ws_hyp.word for ws_hyp in cb_results]), ""

    # step 2: get word-level alignment [word, start_frame, end_frame]
    word_alignment = []
    word = ""
    (
        L,
        r,
    ) = (
        None,
        None,
    )
    for item in alignment_tokens:
        if not word:
            word = item[1][1:]
            L = r = item[0]
        else:
            if item[1].startswith(bow):
                word_alignment.append((word, L, r))
                word = item[1][1:]
                L = r = item[0]
            else:
                word += item[1]
                r = item[0]
    word_alignment.append((word, L, r))
    initial_text_transcript = " ".join([item[0] for item in word_alignment])
    if print_stats:
        logging.info(f"Word alignment: {word_alignment}")

    # step 3: merge spotted words with word alignment
    for ws_hyp in cb_results:
        # extend ws_hyp start frame in case of rnnt (rnnt tends to predict labels one frame earlier sometimes)
        if ws_hyp.start_frame > 0 and decoder_type == "rnnt":
            ws_hyp.start_frame -= 1
        new_word_alignment = []
        already_inserted = False
        # get interval of spotted word
        ws_interval = set(range(ws_hyp.start_frame, ws_hyp.end_frame + 1))
        for item in word_alignment:
            # get interval if word from alignment
            li, ri = item[1], item[2]
            item_interval = set(range(li, ri + 1))
            if ws_hyp.start_frame < li:
                # spotted word starts before first word from alignment
                if not already_inserted:
                    new_word_alignment.append((ws_hyp.word, ws_hyp.start_frame, ws_hyp.end_frame))
                    already_inserted = True
            # compute intersection between spotted word and word from alignment in percentage
            intersection_part = 100 / len(item_interval) * len(ws_interval & item_interval)
            if intersection_part <= intersection_threshold:
                new_word_alignment.append(item)
            elif not already_inserted:
                # word from alignment will be replaced by spotted word
                new_word_alignment.append((ws_hyp.word, ws_hyp.start_frame, ws_hyp.end_frame))
                already_inserted = True
        # insert last spotted word if not yet
        if not already_inserted:
            new_word_alignment.append((ws_hyp.word, ws_hyp.start_frame, ws_hyp.end_frame))
        word_alignment = new_word_alignment
        if print_stats:
            logging.info(f"Spotted word: {ws_hyp.word} [{ws_hyp.start_frame}, {ws_hyp.end_frame}]")

    boosted_text_list = [item[0] for item in new_word_alignment]
    boosted_text = " ".join(boosted_text_list)

    return boosted_text, initial_text_transcript


def compute_fscore(
    recognition_results_manifest: str,
    key_words_list: List,
    eps: str = "<eps>",
    print_stats: bool = False,
) -> tuple[float, float, float]:
    """
    Compute fscore for list of context biasing words/phrases.
    The idea is to get a word-level alignment for ground truth text and prediction results from manifest file.
    Then compute f-score for each word/phrase from key_words_list according to obtained word alignment.

    Args:
        recognition_results_manifest: path to nemo manifest file with recognition results in pred_text field.
        key_words_list: list of context biasing words/phrases.
        return_scores: if True, return precision, recall and fscore (not only print).
        eps: epsilon symbol for alignment.
    Returns:
        Returns tuple of precision, recall and fscore.
    """

    assert key_words_list, "key_words_list is empty"

    # get data from manifest
    assert os.path.isfile(recognition_results_manifest), f"manifest file {recognition_results_manifest} doesn't exist"
    data = read_manifest(recognition_results_manifest)
    assert len(data) > 0, "manifest file is empty"
    assert data[0].get('text', None), "manifest file should contain text field"
    assert data[0].get('pred_text', None), "manifest file should contain pred_text field"

    # compute max number of words in one context biasing phrase
    max_ngram_order = max([len(item.split()) for item in key_words_list])
    key_words_stat = {}  # a word here can be single word or phareses
    for word in key_words_list:
        key_words_stat[word] = [0, 0, 0]  # [true positive (tp), groud truth (gt), false positive (fp)]

    for item in data:
        # get alignment by texterrors
        ref = item['text'].split()
        hyp = item['pred_text'].split()
        ali = align(ref, hyp, eps)

        # 1-grams
        for idx in range(len(ali)):
            word_ref = ali[idx][0]
            word_hyp = ali[idx][1]
            if word_ref in key_words_stat:
                key_words_stat[word_ref][1] += 1  # add to gt
                if word_ref == word_hyp:
                    key_words_stat[word_ref][0] += 1  # add to tp
            elif word_hyp in key_words_stat:
                key_words_stat[word_hyp][2] += 1  # add to fp

        # 2-grams and higher (takes into account epsilons in alignment)
        for ngram_order in range(2, max_ngram_order + 1):
            # for reference phrase
            idx = 0
            item_ref = []
            while idx < len(ali):
                if item_ref:
                    item_ref = [item_ref[1]]
                    idx = item_ref[0][1] + 1  # idex of second non eps word + 1
                while len(item_ref) != ngram_order and idx < len(ali):
                    word = ali[idx][0]
                    idx += 1
                    if word == eps:
                        continue
                    else:
                        item_ref.append((word, idx - 1))
                if len(item_ref) == ngram_order:
                    phrase_ref = " ".join([item[0] for item in item_ref])
                    phrase_hyp = " ".join([ali[item[1]][1] for item in item_ref])
                    if phrase_ref in key_words_stat:
                        key_words_stat[phrase_ref][1] += 1  # add to gt
                        if phrase_ref == phrase_hyp:
                            key_words_stat[phrase_ref][0] += 1  # add to tp
            # in case of false positive hypothesis phrase
            idx = 0
            item_hyp = []
            while idx < len(ali):
                if item_hyp:
                    item_hyp = [item_hyp[1]]
                    idx = item_hyp[0][1] + 1  # idex of first non eps word in previous ngram + 1
                while len(item_hyp) != ngram_order and idx < len(ali):
                    word = ali[idx][1]
                    idx += 1
                    if word == eps:
                        continue
                    else:
                        item_hyp.append((word, idx - 1))
                if len(item_hyp) == ngram_order:
                    phrase_hyp = " ".join([item[0] for item in item_hyp])
                    phrase_ref = " ".join([ali[item[1]][0] for item in item_hyp])
                    if phrase_hyp in key_words_stat and phrase_hyp != phrase_ref:
                        key_words_stat[phrase_hyp][2] += 1  # add to fp

    tp = sum([key_words_stat[x][0] for x in key_words_stat])
    gt = sum([key_words_stat[x][1] for x in key_words_stat])
    fp = sum([key_words_stat[x][2] for x in key_words_stat])

    precision = tp / (tp + fp + 1e-8)
    recall = tp / (gt + 1e-8)
    fscore = 2 * (precision * recall) / (precision + recall + 1e-8)

    if print_stats:
        logging.info("=" * 60)
        logging.info("Per words statistic (word: correct/totall | false positive):\n")
    max_len = max([len(x) for x in key_words_stat if key_words_stat[x][1] > 0 or key_words_stat[x][2] > 0])
    for word in key_words_list:
        if key_words_stat[word][1] > 0 or key_words_stat[word][2] > 0:
            false_positive = ""
            if key_words_stat[word][2] > 0:
                false_positive = key_words_stat[word][2]
            if print_stats:
                logging.info(
                    f"{word:>{max_len}}: {key_words_stat[word][0]:3}/{key_words_stat[word][1]:<3} |{false_positive:>3}"
                )
    if print_stats:
        logging.info("=" * 60)
        logging.info("=" * 60)
        logging.info(f"Precision: {precision:.4f} ({tp}/{tp + fp}) fp:{fp}")
        logging.info(f"Recall:    {recall:.4f} ({tp}/{gt})")
        logging.info(f"Fscore:    {fscore:.4f}")
        logging.info("=" * 60)

    return (precision, recall, fscore)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from lhotse.cut import Cut, MixedCut

from nemo.collections.common.data.prompt_fn import registered_prompt_format_fn
from nemo.collections.common.prompts.formatter import Modality, PromptFormatter


class PlainPromptFormatter(PromptFormatter):
    """
    Plain prompt formatter: there is nothing being added to the user and assistants turns.
    """

    NAME = "plain"
    OUTPUT_ROLE = "assistant"
    TEMPLATE = {
        "user": {
            "template": f"|message|",
            "slots": {
                "message": Modality.Text,
            },
        },
        OUTPUT_ROLE: {
            "template": f"|message|",
            "slots": {
                "message": Modality.Text,
            },
        },
    }


@registered_prompt_format_fn(Cut, PlainPromptFormatter)
def plain(cut: Cut, prompt: PlainPromptFormatter):
    if isinstance(cut, MixedCut):
        cut = cut.first_non_padding_cut
    if cut.has_custom("context"):
        ctx = cut.context
    else:
        ctx = ""

    turns = [{"role": "user", "slots": {"message": ctx}}]
    if (answer := cut.supervisions[0].text) is not None:
        turns.append({"role": "assistant", "slots": {"message": answer}})

    return prompt.encode_dialog(turns)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest
import torch

from nemo.collections.llm.api import finetune
from nemo.collections.vlm import LoRA, MLlamaConfig90BInstruct, MLlamaModel
from nemo.collections.vlm.recipes import mllama_90b
from nemo.lightning import Trainer


class TestMLLama90B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        # Return the module containing the mllama_90b recipe
        return mllama_90b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        # Verify that model() returns a run.Config wrapping MLlamaModel
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == MLlamaModel
        # Verify that the inner configuration is for MLlamaConfig90BInstruct
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == MLlamaConfig90BInstruct

    def test_finetune_recipe_default(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        # Check that the returned recipe is a run.Partial wrapping finetune
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune

        # Verify the model configuration within the recipe
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == MLlamaModel

        # Verify trainer configuration
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert recipe.trainer.accelerator == "gpu"
        # Default values: num_nodes=1 and num_gpus_per_node=8
        assert recipe.trainer.num_nodes == 1
        assert recipe.trainer.devices == 8

        # Check strategy settings from the trainer
        strat = recipe.trainer.strategy
        assert isinstance(strat, run.Config)
        assert strat.tensor_model_parallel_size == 8
        assert strat.pipeline_model_parallel_size == 1
        assert strat.encoder_pipeline_model_parallel_size == 0
        assert strat.pipeline_dtype == torch.bfloat16

        # Verify data configuration
        data = recipe.data
        assert isinstance(data, run.Config)
        # Confirm that the data module is the expected MockDataModule
        assert data.__fn_or_cls__.__name__ == "MockDataModule"
        assert data.seq_length == 6404  # Encoder (vision) sequence length
        assert data.decoder_seq_length == 2048  # Decoder (LLM) sequence length
        assert data.global_batch_size == 16
        assert data.micro_batch_size == 2
        assert data.vocab_size == 128256
        assert data.crop_size == (560, 560)
        assert data.num_workers == 0

        # Check that logging and resume configurations are present
        assert recipe.log is not None
        assert recipe.resume is not None

    def test_finetune_recipe_peft_lora(self, recipe_module):
        # Test with peft_scheme set to "lora"
        recipe = recipe_module.finetune_recipe(peft_scheme="lora")
        # Verify that the recipe contains a peft field configured for LoRA
        assert hasattr(recipe, "peft")
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA

        # Check LoRA-specific parameters
        peft_config = recipe.peft
        assert peft_config.freeze_vision_model is True
        expected_modules = ["linear_qkv", "linear_q", "linear_kv"]
        assert peft_config.target_modules == expected_modules
        assert peft_config.dim == 8
        assert peft_config.alpha == 32
        assert peft_config.dropout == 0.05
        assert peft_config.dropout_position == "pre"

        # The learning rate in the optimizer should be updated for LoRA usage
        assert recipe.optim.config.lr == 1e-4

    def test_invalid_peft_scheme(self, recipe_module):
        # When peft_scheme is 'none', the recipe should raise a ValueError.
        with pytest.raises(ValueError):
            recipe_module.finetune_recipe(peft_scheme="none")

    @pytest.mark.parametrize("num_nodes, num_gpus", [(1, 8), (2, 4)])
    def test_finetune_recipe_different_configurations(self, recipe_module, num_nodes, num_gpus):
        # Validate that the recipe respects different numbers of nodes and GPUs per node
        recipe = recipe_module.finetune_recipe(num_nodes=num_nodes, num_gpus_per_node=num_gpus)
        assert recipe.trainer.num_nodes == num_nodes
        assert recipe.trainer.devices == num_gpus



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import pytest

from nemo.lightning.ckpt_utils import ckpt_to_context_subdir, ckpt_to_dir, idempotent_path_append
from nemo.lightning.io.pl import ckpt_to_weights_subdir


@pytest.mark.parametrize("base_dir", [Path("test/path"), "test/path", "msc://default/tmp/test/path"])
def test_idempotent_path_append_path_no_suffix(base_dir):
    suffix = "new_suffix"
    result = idempotent_path_append(base_dir, suffix)
    assert str(result).endswith("test/path/new_suffix")


@pytest.mark.parametrize(
    "filepath", [Path("test/checkpoints"), "test/checkpoints", "msc://default/tmp/test/checkpoints"]
)
def test_ckpt_to_context_subdir(filepath):
    result = ckpt_to_context_subdir(filepath)
    assert str(result).endswith("test/checkpoints/context")


@pytest.mark.parametrize(
    "filepath", [Path("test/checkpoints"), "test/checkpoints", "msc://default/tmp/test/checkpoints"]
)
def test_ckpt_to_dir(filepath):
    result = ckpt_to_dir(filepath)
    assert str(result).endswith("test/checkpoints")


@pytest.mark.parametrize(
    "filepath", [Path("test/checkpoints"), "test/checkpoints", "msc://default/tmp/test/checkpoints"]
)
def test_ckpt_to_weights_subdir(filepath):
    result = ckpt_to_weights_subdir(filepath, is_saving=True)
    assert str(result).endswith("test/checkpoints/weights")



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import namedtuple
from tqdm import tqdm

from nemo.utils import logging

try:
    import pandas as pd
    from tabulate import tabulate

    HAVE_TABLUATE_AND_PANDAS = True
except (ImportError, ModuleNotFoundError):
    HAVE_TABLUATE_AND_PANDAS = False


def punctuation_error_rate(
    references: list[str], hypotheses: list[str], punctuation_marks: list[str], punctuation_mask: str = "[PUNCT]",
) -> None:

    """
    Computes Punctuation Error Rate
    
    Args:
        references (list[str]) - list of references
        hypotheses (list[str]) - list of hypotheses
        punctuation_marks (list[str]) - list of punctuation marks for computing metrics
        punctuation_mask (str, by default "[PUNCT]") - mask token that will be applied to
        given punctuation marks while edit distance calculation
        
    Return:
        punct_er (float) - Punctuation Error Rate
    """

    dper_obj = DatasetPunctuationErrorRate(
        references=references,
        hypotheses=hypotheses,
        punctuation_marks=punctuation_marks,
        punctuation_mask=punctuation_mask,
    )

    dper_obj.compute()

    return dper_obj.punct_er


class OccurancePunctuationErrorRate:
    """
    Class for computation puncutation-related absolute amounts of operations and thier rates
    between reference and hypothesis strings:
        - Absolute amounts of correct predictions, deletions, insertions
        and substitutions for each given punctuation mark
        - Rates of correct predictions, deletions, insertions
        and substitutions for each given punctuation mark
        - Overall rates of correct predictions, deletions, insertions
        and substiturions between reference and hypothesis string
        - Punctuation Error Rate

    Args to init:
        punctuation_marks (list[str]) - list of punctuation marks for computing metrics
        punctuation_mask (str, by default "[PUNCT]") - mask token that will be applied to
        given punctuation marks while edit distance calculation 
    
    How to use:
        1. Create object of OccurancePunctuationErrorRate class.
           Example:
                punctuation_marks = [".", ",", "!", "?"]
                oper_obj = OccurancePunctuationErrorRate(punctuation_marks)
        
        2. To compute punctuation metrics, pass reference and hypothesis string to the "compute" method
        of created object.
            Example:
                reference_str = "Hi, dear! Nice to see you. What's"
                hypothesis_str = "Hi dear! Nice to see you! What's?"
                oper_obj.compute(reference_str, hypothesis_str)

    Output (listed in order of output):
        1. Dict of absolute operations amounts for each given punctuation mark:
            Example:
            {'.': {'Correct': 0, 'Deletions': 0, 'Insertions': 0, 'Substitutions': 1},
             ',': {'Correct': 0, 'Deletions': 1, 'Insertions': 0, 'Substitutions': 0},
             '!': {'Correct': 1, 'Deletions': 0, 'Insertions': 0, 'Substitutions': 0},
             '?': {'Correct': 0, 'Deletions': 0, 'Insertions': 1, 'Substitutions': 0}}
              
        2. Dict of substitutions absolute amounts between given punctuation marks:
            Example:
            {'.': {'.': 0, ',': 0, '!': 1, '?': 0},
             ',': {'.': 0, ',': 0, '!': 0, '?': 0},
             '!': {'.': 0, ',': 0, '!': 0, '?': 0},
             '?': {'.': 0, ',': 0, '!': 0, '?': 0}}
            
        3. namedtuple "PunctuationRates" of punctuation operation rates (in range from 0 to 1):
            3.1. correct_rate - overall correct rate 
                Example: correct_rate=0.25
            3.2. deletions_rate - overall deletions rate
                Example: deletions_rate=0.25
            3.3. insertions_rate - overall insertions rate
                Example: insertions_rate=0.25
            3.4. substitutions_rate - overall substitutions_rate
                Example: substitutions_rate=0.25
            3.5. punct_er - Punctuation Error Rate
                Example: punct_er=0.75
            3.6. operation_rates - dict of operations rates for each given punctuation mark
                Example: 
                operation_rates={
                    '.': {'Correct': 0.0, 'Deletions': 0.0, 'Insertions': 0.0, 'Substitutions': 1.0},
                    ',': {'Correct': 0.0, 'Deletions': 1.0, 'Insertions': 0.0, 'Substitutions': 0.0},
                    '!': {'Correct': 1.0, 'Deletions': 0.0, 'Insertions': 0.0, 'Substitutions': 0.0},
                    '?': {'Correct': 0.0, 'Deletions': 0.0, 'Insertions': 1.0, 'Substitutions': 0.0}
                    }
  
            3.7. substitution_rates - dict of substitution rates for each given punctuation mark
                Example:
                substitution_rates={
                    '.': {'.': 0.0, ',': 0.0, '!': 1.0, '?': 0.0},
                    ',': {'.': 0.0, ',': 0.0, '!': 0.0, '?': 0.0},
                    '!': {'.': 0.0, ',': 0.0, '!': 0.0, '?': 0.0},
                    '?': {'.': 0.0, ',': 0.0, '!': 0.0, '?': 0.0}
                    }
    """

    def __init__(self, punctuation_marks: list[str], punctuation_mask: str = "[PUNCT]") -> None:

        assert len(punctuation_marks) != 0, f"List of punctuation marks is empty"

        self.punctuation_marks = punctuation_marks
        self.punctuation_mask = punctuation_mask

        self.operations = ["Correct", "Deletions", "Insertions", "Substitutions"]

    def compute_rates(self, operation_amounts: dict, substitution_amounts: dict):
        operation_rates = {pm: {operation: 0 for operation in self.operations} for pm in self.punctuation_marks}
        substitution_rates = {pm: {pm: 0 for pm in self.punctuation_marks} for pm in self.punctuation_marks}

        for pm in self.punctuation_marks:
            operations_amount_by_pm = sum(operation_amounts[pm].values())

            if operations_amount_by_pm == 0:
                continue

            operation_rates[pm] = {
                operation: (operation_amounts[pm][operation] / operations_amount_by_pm)
                for operation in self.operations
            }

            substitution_rates[pm] = {
                _pm: (substitution_amounts[pm][_pm] / operations_amount_by_pm)
                for _pm in substitution_amounts[pm].keys()
            }

        _operation_amounts = {
            operation: {pm: operation_amounts[operation] for pm, operation_amounts in operation_amounts.items()}
            for operation in self.operations
        }

        overall_amounts_by_operation = {
            operation: sum(_operation_amounts[operation].values()) for operation in _operation_amounts
        }
        overall_operations_amount = sum(overall_amounts_by_operation.values())

        punctuation_rates = namedtuple(
            'PunctuationRates',
            [
                'correct_rate',
                'deletions_rate',
                'insertions_rate',
                'substitutions_rate',
                'punct_er',
                'operation_rates',
                'substitution_rates',
            ],
        )

        if overall_operations_amount == 0:
            rates = punctuation_rates(0, 0, 0, 0, 0, operation_rates, substitution_rates)
        else:
            correct_rate = overall_amounts_by_operation["Correct"] / overall_operations_amount
            deletions_rate = overall_amounts_by_operation["Deletions"] / overall_operations_amount
            insertions_rate = overall_amounts_by_operation["Insertions"] / overall_operations_amount
            substitutions_rate = overall_amounts_by_operation["Substitutions"] / overall_operations_amount
            punct_er = deletions_rate + insertions_rate + substitutions_rate

            rates = punctuation_rates(
                correct_rate,
                deletions_rate,
                insertions_rate,
                substitutions_rate,
                punct_er,
                operation_rates,
                substitution_rates,
            )

        return rates

    def compute_operation_amounts(self, reference: str, hypothesis: str):
        operation_amounts = {pm: {operation: 0 for operation in self.operations} for pm in self.punctuation_marks}
        substitution_amounts = {pm: {pm: 0 for pm in self.punctuation_marks} for pm in self.punctuation_marks}

        def tokenize(text: str, punctuation_marks: list[str]):
            punctuation_marks = "\\" + "\\".join(self.punctuation_marks)
            tokens = re.findall(rf"[\w']+|[{punctuation_marks}]", text)
            return tokens

        def mask_punct_tokens(tokens: list[str], punctuation_marks: list[str], punctuation_mask: str):
            masked = [punctuation_mask if token in punctuation_marks else token for token in tokens]
            return masked

        r_tokens = tokenize(reference, self.punctuation_marks)
        h_tokens = tokenize(hypothesis, self.punctuation_marks)

        r_masked = mask_punct_tokens(r_tokens, self.punctuation_marks, self.punctuation_mask)
        h_masked = mask_punct_tokens(h_tokens, self.punctuation_marks, self.punctuation_mask)

        r_punct_amount = r_masked.count(self.punctuation_mask)
        h_punct_amount = h_masked.count(self.punctuation_mask)

        if r_punct_amount + h_punct_amount == 0:
            return operation_amounts, substitution_amounts

        r_len = len(r_masked)
        h_len = len(h_masked)

        costs = [[0 for inner in range(h_len + 1)] for outer in range(r_len + 1)]
        backtrace = [[0 for inner in range(h_len + 1)] for outer in range(r_len + 1)]

        COR = 'C'
        DEL, DEL_PENALTY = 'D', 1
        INS, INS_PENALTY = 'I', 1
        SUB, SUB_PENALTY = 'S', 1

        for i in range(1, r_len + 1):
            costs[i][0] = DEL_PENALTY * i
            backtrace[i][0] = DEL

        for j in range(1, h_len + 1):
            costs[0][j] = INS_PENALTY * j
            backtrace[0][j] = INS

        for j in range(1, h_len + 1):
            costs[0][j] = INS_PENALTY * j
            backtrace[0][j] = INS

        for i in range(1, r_len + 1):
            for j in range(1, h_len + 1):
                if r_masked[i - 1] == h_masked[j - 1]:
                    costs[i][j] = costs[i - 1][j - 1]
                    backtrace[i][j] = COR
                else:
                    substitution_cost = costs[i - 1][j - 1] + SUB_PENALTY
                    insertion_cost = costs[i][j - 1] + INS_PENALTY
                    deletion_cost = costs[i - 1][j] + DEL_PENALTY

                    costs[i][j] = min(substitution_cost, insertion_cost, deletion_cost)
                    if costs[i][j] == substitution_cost:
                        backtrace[i][j] = SUB
                    elif costs[i][j] == insertion_cost:
                        backtrace[i][j] = INS
                    else:
                        backtrace[i][j] = DEL

        i = r_len
        j = h_len

        while i > 0 or j > 0:
            if backtrace[i][j] == COR:
                if r_masked[i - 1] == self.punctuation_mask or h_masked[j - 1] == self.punctuation_mask:
                    r_token = r_tokens[i - 1]
                    h_token = h_tokens[j - 1]

                    if r_token == h_token:
                        operation_amounts[r_token]['Correct'] += 1
                    else:
                        operation_amounts[r_token]['Substitutions'] += 1
                        substitution_amounts[r_token][h_token] += 1
                i -= 1
                j -= 1

            elif backtrace[i][j] == SUB:
                i -= 1
                j -= 1

            elif backtrace[i][j] == INS:
                j -= 1

            elif backtrace[i][j] == DEL:
                i -= 1

        for pm in self.punctuation_marks:
            num_of_correct = operation_amounts[pm]['Correct']

            num_substitutions_of_pm = operation_amounts[pm]['Substitutions']
            num_substitutions_to_pm = sum([substitution_amounts[_pm][pm] for _pm in self.punctuation_marks])

            num_of_deletions = r_tokens.count(pm) - (num_of_correct + num_substitutions_of_pm)
            operation_amounts[pm]['Deletions'] = num_of_deletions

            num_of_insertions = h_tokens.count(pm) - (num_of_correct + num_substitutions_to_pm)
            operation_amounts[pm]['Insertions'] = num_of_insertions

        return operation_amounts, substitution_amounts

    def compute(self, reference: str, hypothesis: str):
        operation_amounts, substitution_amounts = self.compute_operation_amounts(reference, hypothesis)
        punctuation_rates = self.compute_rates(operation_amounts, substitution_amounts)
        return operation_amounts, substitution_amounts, punctuation_rates


class DatasetPunctuationErrorRate:
    """
    Class for computation the total puncutation-related absolute amounts of operations and their rates 
    in pairs of reference and hypothesis strins:
        - Absolute amounts of correct predictions, deletions, insertions
        and substitutions for each given punctuation mark
        - Rates of correct predictions, deletions, insertions
        and substitutions for each given punctuation mark 
        - Total rates of correct predictions, deletions, insertions
        and substiturions in pairs of reference and hypothesis strings 
        - Punctuation Error Rate
        
    Args to init:
        references (list[str]) - list of references
        hypotheses (list[str]) - list of hypotheses
        punctuation_marks (list[str]) - list of punctuation marks for computing metrics
        punctuation_mask (str, by default "[PUNCT]") - mask token that will be applied to
        given punctuation marks while edit distance calculation
        
    How to use:
        1. Create object of DatasetPunctuationErrorRate class.
           Example:
                references = ["Hi, dear! Nice to see you. What's"]
                hypotheses = ["Hi dear! Nice to see you! What's?"]                
                punctuation_marks = [".", ",", "!", "?"]
                
                dper_obj = DatasetPunctuationErrorRate(references, hypotheses, punctuation_marks)
                
        2. To compute punctuation metrics, call the class method "compute()".
            Example:
                dper_obj.compute() 
                
    Result:
    The following atributes of class object will be updated with calculated metrics values.
    The values are available with calling the atributes:
        
        dper_obj.operation_rates - dict, rates of correctness and errors for each punctuation mark 
        from `preset dper_obj.punctuation_marks` list.
        
        dper_obj.substitution_rates - dict, substitution rates between puncutation marks from
        `preset dper_obj.punctuation_marks` list.
        
        dper_obj.correct_rate - float, total rate of correctness between provided pairs of 
        references and hypotheses.
        
        dper_obj.deletions_rate - float, total rate of deletions between provided pairs of 
        references and hypotheses.
        
        dper_obj.insertions_rate - float, total rate of insertions between provided pairs of 
        references and hypotheses.
        
        dper_obj.substitutions_rate - float, total rate of substitutions between provided pairs of 
        references and hypotheses.
        
        dper_obj.punct_er - float, total Punctuation Error Rate between provided pairs of 
        references and hypotheses.
    """

    def __init__(
        self,
        references: list[str],
        hypotheses: list[str],
        punctuation_marks: list[str],
        punctuation_mask: str = "[PUNCT]",
    ) -> None:

        self.references = references
        self.hypotheses = hypotheses
        self.punctuation_marks = punctuation_marks
        self.punctuation_mask = punctuation_mask

        self.oper_obj = OccurancePunctuationErrorRate(
            punctuation_marks=self.punctuation_marks, punctuation_mask=self.punctuation_mask
        )

        self.operation_amounts = []
        self.substitution_amounts = []
        self.rates = []

        self.operation_rates = None
        self.substitution_rates = None
        self.correct_rate = None
        self.deletions_rate = None
        self.insertions_rate = None
        self.substitutions_rate = None
        self.punct_er = None

    def compute(self):
        def sum_amounts(amounts_dicts: list[dict]):
            amounts = {key: {_key: 0 for _key in amounts_dicts[0][key]} for key in amounts_dicts[0].keys()}

            for amounts_dict in amounts_dicts:
                for outer_key, inner_dict in amounts_dict.items():
                    for inner_key, value in inner_dict.items():
                        amounts[outer_key][inner_key] += value
            return amounts

        logging.info("Computing Punctuation Error Rate")

        for reference, hypothesis in tqdm(zip(self.references, self.hypotheses), total=len(self.references)):
            operation_amounts, substitution_amounts, punctuation_rates = self.oper_obj.compute(reference, hypothesis)
            self.operation_amounts.append(operation_amounts)
            self.substitution_amounts.append(substitution_amounts)
            self.rates.append(punctuation_rates)

        overall_operation_amounts = sum_amounts(self.operation_amounts)
        overall_substitution_amounts = sum_amounts(self.substitution_amounts)
        overall_rates = self.oper_obj.compute_rates(
            operation_amounts=overall_operation_amounts, substitution_amounts=overall_substitution_amounts
        )

        self.operation_rates = overall_rates.operation_rates
        self.substitution_rates = overall_rates.substitution_rates
        self.correct_rate = overall_rates.correct_rate
        self.deletions_rate = overall_rates.deletions_rate
        self.insertions_rate = overall_rates.insertions_rate
        self.substitutions_rate = overall_rates.substitutions_rate
        self.punct_er = overall_rates.punct_er

    def reset(self):
        self.operation_amounts = []
        self.substitution_amounts = []
        self.rates = []

        self.operation_rates = None
        self.substitution_rates = None
        self.correct_rate = None
        self.deletions_rate = None
        self.insertions_rate = None
        self.substitutions_rate = None
        self.punct_er = None

    def print(self):
        logging.info(f'Dataset PER ' + str(round(100 * self.punct_er, 2)) + '%')

        if HAVE_TABLUATE_AND_PANDAS:
            rates_by_pm_df = pd.DataFrame(self.operation_rates) * 100
            substitution_rates_by_pm_df = pd.DataFrame(self.substitution_rates) * 100

            logging.info(
                "Rates of punctuation correctness and errors (%):\n"
                + tabulate(rates_by_pm_df, headers='keys', tablefmt='psql')
            )
            logging.info(
                "Substitution rates between punctuation marks (%):\n"
                + tabulate(substitution_rates_by_pm_df, headers='keys', tablefmt='psql')
            )
        else:
            logging.warning("Some of the modules (pandas or tabulate) can't be imported")
            logging.info(f"Rates of punctuation correctness and errors (in range [0, 1]):\n{self.operation_rates}\n")
            logging.info(
                f"Substitution rates between punctuation marks (in range [0, 1]):\n{self.substitution_rates}\n"
            )



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import signal
import sys
from typing import Optional

import torch
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.trainer.trainer import Trainer

from nemo.lightning.io.mixin import IOMixin
from nemo.utils import logging
from nemo.utils.callbacks.dist_ckpt_io import AsyncFinalizableCheckpointIO


class PreemptionCallback(Callback, IOMixin):
    """
    PreemptionCallback checks for preemption during training at the end of every step.
    Upon preemption, it signals the trainer to stop gracefully.

    Args:
        sig (int, optional): The signal to listen for. Defaults to signal.SIGTERM.

    Example:
        >>> from nemo.lightning.pytorch.callbacks import PreemptionCallback
        >>> callback = PreemptionCallback()
        >>> trainer = Trainer(callbacks=[callback])
    """

    def __init__(self, sig: Optional[int] = None):
        self.sig = sig if sig is not None else signal.SIGTERM
        self._interrupted = False
        self._handler_context = None
        self._preemption_supported = None

    def on_train_start(self, trainer: Trainer, pl_module) -> None:
        if self.preemption_supported:
            self._handler_context = self._preemption_handler()
            self._handler_context.__enter__()

    def on_train_batch_start(self, trainer: Trainer, pl_module, batch, batch_idx: int) -> None:
        if not self.preemption_supported:
            self._preemption_supported = self._check_preemption_support()
            if self.preemption_supported:
                self._handler_context = self._preemption_handler()
                self._handler_context.__enter__()

    def on_train_end(self, trainer: Trainer, pl_module) -> None:
        if self._handler_context:
            self._handler_context.__exit__(None, None, None)

    def on_train_batch_end(self, trainer: Trainer, pl_module, outputs, batch, batch_idx: int) -> None:
        if self.interrupted:
            logging.info("Preemption detected, saving checkpoint and exiting")
            trainer.should_stop = True
            if trainer.checkpoint_callback:
                monitor_candidates = trainer.checkpoint_callback._monitor_candidates(trainer)
                trainer.checkpoint_callback._save_last_checkpoint(trainer, monitor_candidates)
                if isinstance(trainer.strategy.checkpoint_io, AsyncFinalizableCheckpointIO):
                    logging.info("Async checkpointing detected, waiting for it to complete")
                    trainer.strategy.checkpoint_io.maybe_finalize_save_checkpoint(blocking=True)
                sys.exit(0)

    @contextlib.contextmanager
    def _preemption_handler(self):
        if not self.preemption_supported:
            logging.warning("Preemption requires torch distributed to be initialized, preemption may be disabled")
            yield
            return

        original_handler = signal.getsignal(self.sig)

        def master_handler(signum, frame):
            logging.info(f"Received signal {signum}, initiating graceful stop")
            self._interrupted = True

        def ignoring_handler(signum, frame):
            logging.debug(f"Received signal {signum} on non-master rank, ignoring")

        try:
            private_rank = torch.distributed.get_rank()
            signal.signal(self.sig, master_handler if private_rank == 0 else ignoring_handler)
            yield
        finally:
            signal.signal(self.sig, original_handler)

    @property
    def preemption_supported(self) -> bool:
        if self._preemption_supported is None:
            self._preemption_supported = self._check_preemption_support()
        return self._preemption_supported

    def _check_preemption_support(self) -> bool:
        return torch.distributed.is_available() and torch.distributed.is_initialized()

    @property
    def interrupted(self) -> bool:
        if not self.preemption_supported:
            return False
        interrupted = torch.tensor(self._interrupted, device=torch.cuda.current_device(), dtype=torch.int32)
        torch.distributed.broadcast(interrupted, 0)
        return bool(interrupted.item())
